# Copyright 2021 National Technology & Engineering Solutions of Sandia, LLC (NTESS).
# Under the terms of Contract DE-NA0003525 with NTESS, the U.S. Government retains
# certain rights in this software.


"""Tools for working with computer vision data stored using the :ref:`specification`."""

__version__ = "0.1.0"
